;; Generated package description from .el  -*- rpm-spec-modeno-byte-compile: t -*-
(define-package "rpm-spec-mode" "0.16.0.20250329.13938" "RPM spec mode for Emacs/XEmacs" '((emacs "27.1")) :commit "8cd329b78c7bc6285b7b9f2c65a58a9e778a59ca" :authors '((nil . "stig@bjorlykke.org") ("Tore Olsen" . "toreo@tihlde.org") ("Steve Sanbeg" . "sanbeg@dset.com") ("Tim Powers" . "timp@redhat.com") ("Trond Eivind Glomsrød" . "teg@redhat.com") ("Chmouel Boudjnah" . "chmouel@mandrakesoft.com") ("Ville Skyttä" . "ville.skytta@iki.fi") ("Adam Spiers" . "elisp@adamspiers.org")) :maintainer '("Björn Bidar" . "bjorn.bidar@thaodan.de") :keywords '("unix" "languages" "rpm") :url "https://github.com/Thaodan/rpm-spec-mode/")
